/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    control::ArmarXObjects::controller_creator
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#pragma once

#include <memory>

#include <Eigen/Dense>

#include <RobotAPI/libraries/core/FramedPose.h>
#include <RobotAPI/interface/units/ForceTorqueUnit.h>
#include <ArmarXCore/observers/variant/DatafieldRef.h>

namespace armarx::control::common::ft
{
    class ForceTorqueUtility;
    using ForceTorqueUtilityPtr = std::shared_ptr<ForceTorqueUtility>;

    class ForceTorqueUtility
    {
    public:
        ForceTorqueUtility(const ForceTorqueUnitObserverInterfacePrx&  forceTorqueObserver, const std::string& FTDatefieldName);

        Eigen::Vector3f getForce();
        FramedDirection getFramedForce();

        Eigen::Vector3f getTorque();
        Eigen::Vector6f getForceTorque();

        void calibrate();

        DatafieldRefPtr forceDf;
        DatafieldRefPtr torqueDf;
        Eigen::Vector3f initialForce = Eigen::Vector3f::Zero();
        Eigen::Vector3f initialTorque = Eigen::Vector3f::Zero();
    };
}
