/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#pragma once


#include <cmath>
#include <memory>
#include <type_traits>

#include <SimoxUtility/math/periodic/periodic_clamp.h>

#include <ArmarXCore/core/rapidxml/wrapper/DefaultRapidXmlReader.h>

// Just so that these functionalities are available
#include <RobotAPI/components/units/RobotUnit/util/RtLogging.h>
#include <RobotAPI/components/units/RobotUnit/util/RtTiming.h>

#include <armarx/control/hardware_config/Types.h>

#include "LinearConvertedValue.h"
#include "ModularConvertedValue.h"

namespace armarx::control::ethercat
{

    /**
     * @class DataInterface
     * @ingroup Library-ethercat
     * @brief Brief description of class DataInterface.
     *
     * Detailed description of class DataInterface.
     */
    class DataInterface
    {

    public:
        DataInterface();
        virtual ~DataInterface();
        virtual void rtReadSensorValues(const IceUtil::Time& sensorValuesTimestamp,
                                        const IceUtil::Time& timeSinceLastIteration) = 0;
        virtual void rtWriteTargetValues(const IceUtil::Time& sensorValuesTimestamp,
                                         const IceUtil::Time& timeSinceLastIteration) = 0;
    };

} // namespace armarx::control::ethercat
