#pragma once

#include <RobotAPI/interface/units/KinematicUnitInterface.h>
#include <RobotAPI/libraries/armem/client/MemoryNameSystem.h>
#include <RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h>
#include <RobotAPI/libraries/skills/provider/SimpleSpecializedSkill.h>

#include <armarx/control/skills/skills/aron/MoveJointsToPositionParams.aron.generated.h>

#include "../skills.h"

namespace armarx::control::skills::skills
{

    /**
   * @class MoveJoints
   * @brief MoveJoints-skill to move joints to specified target values using the KinematicUnit
   *
   * Sets joint values with the KinematicUnit and waits until target values are reached. Can
   * simulate trajectory when in simulation (otherwise trajectories from joint value updates through
   * the KinematicUnit are not simulated).
   */
    class MoveJointsToPosition :
        public ::armarx::skills::SimpleSpecializedSkill<MoveJointsToPositionParams>
    {

    public:
        struct Remote : public armarx::control::skills::KinematicUnitRemote
        {
            ::armarx::armem::robot_state::VirtualRobotReader virtualRobotReader;
            std::string robotName;
        };

        static ::armarx::skills::SkillDescription
        GetSkillDescription()
        {
            return ::armarx::skills::SkillDescription{
                .skillId = {.skillName = "MoveJointsToPosition"},
                .description = "Moves joints to specified target values using the KinematicUnit",
                .timeout = ::armarx::Duration::Minutes(1),
                .parametersType = ParamType::ToAronType()};
        }

        MoveJointsToPosition(const Remote&);


    private:
        ::armarx::skills::Skill::MainResult main(const SpecializedMainInput& in) override;
        ::armarx::skills::Skill::ExitResult exit(const SpecializedExitInput& in) override;

        std::map<std::string, float>
        filterJointTargetValues(const NameValueMap& jointsTargetValues,
                                const std::vector<std::string>& disabledJoints) const;

        bool checkJointsTargetValuesValid(const NameValueMap& jointsTargetValues,
                                          const VirtualRobot::RobotPtr& virtualRobot);

        void switchControlMode(const NameValueMap& jointsTargetValues,
                               const armarx::ControlMode controlMode);

        void moveJoints(const NameValueMap& jointsTargetValues,
                        float accuracyDefault,
                        const NameValueMap& accuracyOverride,
                        const VirtualRobot::RobotPtr& virtualRobot);

        void moveJointsSimulation(const NameValueMap& jointsTargetValues,
                                  float speedDefault,
                                  const NameValueMap& speedOverride,
                                  const VirtualRobot::RobotPtr& virtualRobot);

    private:
        Remote remote;
    };

} // namespace armarx::control::skills::skills
